import scipy
import skimage.io as skio
import skimage.transform as sktransform
import numpy as np


class MultifractalVBMs:
    """
    A class that can perform multifractal biomarker computation for a fundus vasculature segmentation.


    :param n_dim: Maximum dimension used to calculate Singularity Length.
    :type n_dim: int

    :param n_rotations: Number of rotations performed on the given segmentation from which the optimal run will be chosen.
    :type n_rotations: int

    :param optimize: A boolean value that specify if the computation will be made using several rotated version of the segmentation.
    :type optimize: bool

    :param min_proba: A minimum probability for the occupancy of calculated boxes (boxes with smaller probability will be ignored).
    :type min_proba: float

    :param max_proba: A maximum probability for the occupancy of calculated boxes (boxes with higher probability will be ignored).
    :type max_proba: float
    """
    def __init__(self, n_dim=10, n_rotations=25, optimize=True, min_proba=0.01, maxproba=0.98):
        """
        Constructor

        
        :param n_dim: Maximum dimension used to calculate Singularity Length.
        :type n_dim: int

        :param n_rotations: Number of rotations performed on the given segmentation from which the optimal run will be chosen.
        :type n_rotations: int

        :param optimize: A boolean value that specify if the computation will be made using several rotated version of the segmentation.
        :type optimize: bool

        :param min_proba: A minimum probability for the occupancy of calculated boxes (boxes with smaller probability will be ignored).
        :type min_proba: float

        :param max_proba: A maximum probability for the occupancy of calculated boxes (boxes with higher probability will be ignored).
        :type max_proba: float

        """
        self.n_dim = n_dim
        self.n_rotations = n_rotations
        self.optimize = optimize
        self.box_occupancy_proba = [min_proba, maxproba]

    def compute_multifractals(self, segmentation):
        '''
        Computes the multifractal biomarkers of a given retinal vasculature segmentation, specifically the dimensions D0, D1, D2, and the singularity length.

        :param segmentation: An (N,N) binary numpy array representing the segmentation to be analyzed.
        :type segmentation: np.ndarray
        :return: A numpy array containing the computed biomarkers [D0, D1, D2, Singularity Length].
        :rtype: np.ndarray
        '''

        rotations = self.n_rotations
        if not self.optimize:
            rotations = 1

        image_stats = self.get_q_fractals(segmentation, rotations=rotations, dim_list=[0, 1, 2, -self.n_dim, self.n_dim])
        first_three_dqs = image_stats[:, :3, 0]

        if self.optimize:
            best_rotation_index = self._optimize_dqs(first_three_dqs)
            image_stats_opt = image_stats[best_rotation_index, :, :]
        else:
            image_stats_opt = image_stats.squeeze()

        sing_features = self.get_singularity_length(alpha=image_stats_opt[-2:, 3])
        three_dqs = image_stats_opt[:3, 0]
        return np.hstack((three_dqs, sing_features))

    def get_singularity_length(self, alpha):
        '''
         This method is used to compute singularity length from f(alpha) singularity curve.

        :param alpha: 1D numpy array containing f(alpha) singularity curve values.
        :type alpha: int
        :return: Singularity length of this f(alpha) curve.
        :rtype: float
        '''
        s_len = alpha.max() - alpha.min()
        return s_len

    def get_multi_fractal_dimension(self, segmentation, q):
        '''
        Calculates Q dimension from the given binary segmentation.

        :param segmentation: (N,N) binary numpy array segmentation.
        :type segmentation: np.ndarray
        :param q: q power of multi-fractal D_q which will be calculated from the segmentation.
        :type q: float
        :return: 4-tuple of Multi-fractal calculations:
            * Dq_slope - D_q value for given q.
            * dq_r2 - R^2 of the fit of D_q.
            * fq_slope - f(q) value for given q.
            * alphaq_slope - \alpha(q) value for given q.
        '''
        # Only for 2d binary segmentation
        assert len(segmentation.shape) == 2
        assert (segmentation.max() == 1) and (segmentation.min() == 0)

        # Minimal dimension of segmentation
        p = min(segmentation.shape)
        # create a linear sampling for epsilons
        epsilons = np.linspace(5, int(0.6 * p), 20).astype(np.int64)

        counts = []
        fq_numerators = []
        alpha_q_numerators = []
        for size in epsilons:
            P = self._probability_per_pixel(segmentation, size, occupancy=self.box_occupancy_proba)
            # If all pixels were filtered, try again with no limit on pixel occupancy
            if np.sum(P) == 0:
                P = self._probability_per_pixel(segmentation, size, occupancy=[0.001, 1])

            assert np.linalg.norm(np.sum(P) - 1) <= 0.000001, 'All probabilities must sum to 1'
            p_positive = P[P > 0]
            I = np.sum(p_positive ** q)

            mu = (p_positive ** q) / I
            fq_numerator = np.sum(mu * np.log(mu))
            alpha_q_numerator = np.sum(mu * np.log(p_positive))

            if q == 1:
                dq = - np.sum(p_positive * np.log(p_positive))
            else:
                dq = np.log(I) / (1 - q)

            counts.append(dq)
            fq_numerators.append(fq_numerator)
            alpha_q_numerators.append(alpha_q_numerator)

        # Fit D(q) - the successive log(sizes) with log (counts)
        dq_slope, _, r_value, _, _ = scipy.stats.linregress(-np.log(epsilons), np.array(counts))
        dq_r2 = r_value ** 2
        # Fit f(q)
        fq_slope, _, _, _, _ = scipy.stats.linregress(np.log(epsilons), np.array(fq_numerators))
        # Fit alpha(q)
        alphaq_slope, _, _, _, _ = scipy.stats.linregress(np.log(epsilons), np.array(alpha_q_numerators))

        return dq_slope, dq_r2, fq_slope, alphaq_slope

    def get_q_fractals(self, segmentation, rotations, dim_list):
        '''
        Calculates all multi-fractal dimensions specified in dim_list for rotation number of times.

        :param : (N,N) binary numpy array segmentation.
        :type segmentation: np.ndarray
        :param rotations: number of rotations performed on the given segmentation from which the optimal run will be chosen.
        :type rotations: int
        :param dim_list: Iterable of dimensions to calculate from segmentation.
        :type dim_list: list
        :return: The output of 'get_multi_fractal_dimension' for every rotation and q dimension.
        :rtype: np.ndarray with size (rotations, len(dim_list), 4).
        '''
        angles = np.linspace(0, 360, rotations)

        dqs = []
        for angle in angles:
            rotated_image = sktransform.rotate(segmentation, angle, resize=True, cval=0, mode='constant')
            dq = []
            for q in dim_list:
                dq_value = self.get_multi_fractal_dimension(rotated_image, q)
                dq.append(dq_value)

            dqs.append(dq)

        dqs = np.array(dqs)

        return dqs

    @staticmethod
    def custom_add_reduceat(segmentation, idx, axis=0):
        '''
        Custom numpy.add.reduceat function, twice as fast in the context of binary segmentations.
        Sum a 2d-segmentation along a given axis, in the intervals specified by 'idx'.

        :param segmentation: (N,N) numpy array.
        :type segmentation: np.ndarray
        :param idx: indices along the axis between which the function should sum the values.
        :type idx: np.ndarray
        :param axis: Specifies the dimension of im on which to operate. Must be {0, 1}.
        :type axis: int
        :return: Aggregated segmentation along the given axis.
        :rtype: np.ndarray with size (idx.shape[0], N) for axis=0 or (N, idx.shape[0]) for axis=1.
        '''

        idxs_ext = np.concatenate((idx, np.array([segmentation.shape[axis]])))
        if axis == 0:
            results = np.empty(shape=(idx.size, segmentation.shape[1]))
            for i in range(idx.size):
                results[i, :] = segmentation[idxs_ext[i]:idxs_ext[i + 1], :].sum(axis=axis)
        elif axis == 1:
            results = np.empty(shape=(segmentation.shape[0], idx.size))
            for j in range(idx.size):
                results[:, j] = segmentation[:, idxs_ext[j]:idxs_ext[j + 1]].sum(axis=axis)
        else:
            raise Exception("Axis must be 0 or 1.")

        return results

    def _probability_per_pixel(self, segmentation, k, occupancy):
        '''
        Creates a grid with size k on the segmentation and calculates the occupancy probability for every such box.
        Filters the boxes to be between occupancy[0] <= pixel <= occupancy[1]

        :param segmentation: (N,N) numpy binary array.
        :type segmentation: np.ndarray
        :param k: Grid box size.
        :type k: int
        :param occupancy: a list containing the [min_proba, max_proba]. Boxes with probability smaller than min_proba and larger than max_proba will be ignored.
        :type occupancy: list[float]
        :returns: A probability grid calculated from segmentation with box size k.
        :rtype: np.ndarray with size (math.ceil(N/k), math.ceil(N/k)).
        '''
        # split segmentation to boxes sized k by two steps: first sum along vertical axis and then along horizontal.
        vertical_sum = self.custom_add_reduceat(segmentation, np.arange(0, segmentation.shape[0], k), axis=0)
        horizontal_sum = self.custom_add_reduceat(vertical_sum, np.arange(0, segmentation.shape[1], k), axis=1)
        M = horizontal_sum

        # calculate probability per box
        p = M / (k ** 2)
        # filter by box occupancy
        condition = (p >= occupancy[0]) & (p <= occupancy[1])
        P = M[condition] / np.sum(M[condition])
        return P

    def _optimize_dqs(self, dqs):
        '''
        Finds the best sampling index among all rotations: 
            a sample which satisfies D0>D1>D2 where D0 is the largest.
            if no sampling satisfies D0>D1>D2 returns the sampling index with largest D0.

        :param dqs: ndarray with dqs.shape = (rand_rotations, 3) with the values of D0,D1,D2 for every rotation.
        :type dqs: np.ndarray
        :returns: Finds the best sampling index among all rotations.
        :rtype: int
        '''

        index1 = (dqs[:, 0] - dqs[:, 1]) > 0
        index2 = (dqs[:, 1] - dqs[:, 2]) > 0

        comb_ind = index1 * index2
        full_indx_arr = np.arange(0, dqs.shape[0], 1)
        dqs_subset = dqs[comb_ind, 0]
        subset_index_arr = full_indx_arr[comb_ind]
        if subset_index_arr.size != 0:
            subset_max_index = np.argmax(dqs_subset)
            global_max_index = subset_index_arr[subset_max_index]
        else:
            global_max_index = np.argmax(dqs[:, 0])

        return global_max_index




import onnxruntime as ort
import numpy as np
import PIL
from PIL import Image
from torchvision import transforms
import os
import cv2

class DiscSegmenter:
    """A class that performs optic disc segmentation."""

    def __init__(self):
        """Initialize the DiscSegmenter class with image size and model path."""
        self.img_size = 512
        script_dir = os.path.dirname(os.path.abspath(__file__))
        self.model_path = os.path.join(script_dir, "lunetv2_odc.onnx")

    def find_biggest_contour(self, contours):
        """
        Find the biggest contour in the provided list of contours.

        :param contours: List of contours.
        :type contours: list of numpy arrays
        :return: The center, radius, and the biggest contour.
        :rtype: tuple
        """
        radius = -1
        final_contour = contours[0]
        (x, y), radius = cv2.minEnclosingCircle(final_contour)
        center = (int(x), int(y))
        radius = int(radius)
        for contour in contours:
            (x, y), radius_ = cv2.minEnclosingCircle(contour)
            if radius_ > radius:
                center = (int(x), int(y))
                radius = int(radius_)
                final_contour = contour
        return center, radius, final_contour

    def post_processing(self, segmentation, max_roi_size):
        """
        Post-process the segmentation result to extract relevant zones.

        :param segmentation: Segmentation result as a numpy array.
        :type segmentation: numpy array
        :param max_roi_size: Maximum size of the region of interest.
        :type max_roi_size: int
        :return: The center, radius, region of interest, and zones ABC.
        :rtype: tuple
        """
        segmentation = np.array(segmentation, dtype=np.uint8)
        try:
            contours, hierarchy = cv2.findContours(segmentation, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
            center, radius, contour = self.find_biggest_contour(contours)

            one_radius = np.zeros((segmentation.shape[0], segmentation.shape[1], 3), dtype=np.uint8)
            cv2.circle(one_radius, center, radius, (0, 255, 0), -1)

            two_radius = np.zeros((segmentation.shape[0], segmentation.shape[1], 3), dtype=np.uint8)
            cv2.circle(two_radius, center, int(radius * 2), (0, 255, 0), -1)

            three_radius = np.zeros((segmentation.shape[0], segmentation.shape[1], 3), dtype=np.uint8)
            cv2.circle(three_radius, center, radius * 3, (0, 255, 0), -1)

            roi = np.zeros((segmentation.shape[0], segmentation.shape[1], 3), dtype=np.uint8)
            cv2.circle(roi, center, max_roi_size, (0, 255, 0), -1)

        except:
            one_radius = np.zeros((segmentation.shape[0], segmentation.shape[1], 3), dtype=np.uint8)
            two_radius = np.zeros((segmentation.shape[0], segmentation.shape[1], 3), dtype=np.uint8)
            three_radius = np.zeros((segmentation.shape[0], segmentation.shape[1], 3), dtype=np.uint8)
            roi = np.zeros((segmentation.shape[0], segmentation.shape[1], 3), dtype=np.uint8)
            center = (segmentation.shape[0] // 2, segmentation.shape[1] // 2)
            radius = 0

        zones_ABC = np.zeros((segmentation.shape[0], segmentation.shape[1], 4))
        zones_ABC[:, :, :3] = one_radius
        zones_ABC[:, :, 0] = two_radius[:, :, 1]
        zones_ABC[:, :, 2] = three_radius[:, :, 1]
        zones_ABC[:, :, 3] = np.maximum(one_radius[:, :, 1], three_radius[:, :, 1]) / 2

        return center, radius, roi, zones_ABC

    def segment(self, image_path):
        """
        Perform the optic disc segmentation given an image path.

        :param image_path: Path to the image.
        :type image_path: str
        :return: A PIL Image containing the Optic Disc segmentation.
        :rtype: PIL.Image
        """
        session = ort.InferenceSession(self.model_path)
        input_name = session.get_inputs()[0].name

        img_orig = PIL.Image.open(image_path)
        original_size = img_orig.size
        image = img_orig.resize((self.img_size, self.img_size))
        image = transforms.ToTensor()(image)
        normalize = transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])
        image = normalize(image)

        image_np = image.numpy()
        image_np = np.expand_dims(image_np, axis=0)

        outputs = session.run(None, {input_name: image_np})
        od = outputs[0][0, 0] > 0

        return PIL.Image.fromarray(np.array(od, dtype=np.uint8) * 255).resize(original_size, PIL.Image.Resampling.NEAREST)



import numpy as np
from skimage.morphology import skeletonize
from scipy.signal import convolve2d
from PVBM.helpers.tortuosity import compute_tortuosity
from PVBM.helpers.perimeter import compute_perimeter_
from PVBM.helpers.branching_angle import compute_angles_dictionary
#from PVBM.helpers.far import far
from PVBM.GraphRegularisation.GraphRegularisation import TreeReg
from PVBM.GraphCRE.GraphCentralRetinalEquivalent import Tree
class CREVBMs:
    """A class that can perform geometrical biomarker computation for a fundus image.
    """

    def extract_subgraphs(self,graphs, x_c, y_c):
        """
        Extract B, the a graph where each of the disconnected subgraph is labeled differently and D which contains the euclidian distance graph between each
        point in the graph and the optic disc.

        :param graphs: Original blood vessel segmentation graph
        :type graphs: array
        :param x_c: x axis of the optic disc center
        :type x_c: int
        :param y_c: y axis of the optic disc center
        :type y_c: int

        :return: B,D
        :rtype: tuple
        """
        B = np.zeros_like(graphs)
        D = np.zeros_like(graphs)
        n = 1
        for i in range(graphs.shape[0]):
            for j in range(graphs.shape[1]):
                if B[i, j] == 0 and graphs[i, j] == 1:
                    self.recursive_subgraph(graphs, B, D, i, j, n, x_c, y_c)
                    n += 1
        return B, D

    def recursive_subgraph(self,A, B, D, i, j, n, x_c, y_c):
        """
        Recursively extract the value within B and D.

        :param A: Original blood vessel segmentation graph
        :type A: array
        :param B: A graph where each of the disconnected subgraph is labeled differentl, which is initialized by a zeros matrix and recursively built
        :type B: array
        :param D: Euclidian Distance graph between each point in A and the optic disc center (x_c,y_c), which is initialized by a zeros matrix and recursively built
        :type D: array
        :param i: Current x axis location within the graph
        :type i: int
        :param j: Current y axis location within the graph
        :type j: int
        :param n: Current number of point distance since the optic disc
        :type n: int
        :param x_c: x axis of the optic disc center
        :type x_c: int
        :param y_c: y axis of the optic disc center
        :type y_c: int

        :return: B,D
        :rtype: tuple
        """
        up = (i - 1, j)
        down = (i + 1, j)
        left = (i, j - 1)
        right = (i, j + 1)

        up_left = (i - 1, j - 1)
        up_right = (i - 1, j + 1)
        down_left = (i + 1, j - 1)
        down_right = (i + 1, j + 1)
        points = [up, down, left, right, up_left, up_right, down_left, down_right]
        for point in points:
            if point[0] >= 0 and point[0] < B.shape[0] and point[1] < B.shape[1] and point[1] >= 0:
                if A[point] == 1:
                    B[point] = n
                    A[point] = 0
                    D[point] = ((y_c - point[0]) ** 2 + (x_c - point[1]) ** 2) ** 0.5
                    self.recursive_subgraph(A, B, D, point[0], point[1], n, x_c, y_c)

    def central_equivalent(self, l,formula):
        """
        Recursively compute the central retinal equivalent given a list of blood vessel measurement and a formula to

        :param l: list of blood vessel measurements
        :type l: List
        :param formula: One of the following formulas: PVBM.CentralRetinalAnalysis.crae_hubbard, PVBM.CentralRetinalAnalysis.crve_hubbard, PVBM.CentralRetinalAnalysis.crae_knudtson or PVBM.CentralRetinalAnalysis.crve_knudtson
        :type formula: python function

        :return: the computed central retinal equivalent VBM
        :rtype: float
        """
        l = np.sort(l[:])
        new_l = []
        pivot =None
        if len(l) == 1:
            return l[0]
        if len(l)%2 !=0:
            idx_pivot = len(l)//2
            pivot =  l[idx_pivot]
        for i in range(len(l)//2):
            new_l.append(formula(l[i],l[-i-1]))
        if pivot != None:
            new_l.append(pivot)
        return self.central_equivalent(new_l,formula)

    def crae_hubbard(self,w1, w2):
        """
        Recursive formulas of the Central Retinal Arteriolar Equivalent according to Hubbard et al.
        :param w1: w1 vessel measurement
        :type w1: float
        :param w2: w2 vessel measurement
        :type w2: float

        :return: the root vessel of w1 and w2 estimation
        :rtype: float
        """
        return (0.87 * w1 ** 2 + 1.01 * w2 ** 2 - 0.22 * w1 * w2 - 10.76) ** 0.5

    def crve_hubbard(self,w1, w2):
        """
        Recursive formulas of the Central Retinal Venular Equivalent according to Hubbard et al.
        :param w1: w1 vessel measurement
        :type w1: float
        :param w2: w2 vessel measurement
        :type w2: float

        :return: the root vessel of w1 and w2 estimation
        :rtype: float
        """
        return (0.72 * w1 ** 2 + 0.91 * w2 ** 2 + 450.05) ** 0.5

    def crae_knudtson(self,w1, w2):
        """
        Recursive formulas of the Central Retinal Arteriolar Equivalent according to Knudtson et al.
        :param w1: w1 vessel measurement
        :type w1: float
        :param w2: w2 vessel measurement
        :type w2: float

        :return: the root vessel of w1 and w2 estimation
        :rtype: float
        """
        return 0.88 * (w1 ** 2 + w2 ** 2) ** 0.5

    def crve_knudtson(self,w1, w2):
        """
        Recursive formulas of the Central Retinal Venular Equivalent according to Knudtson et al.
        :param w1: w1 vessel measurement
        :type w1: float
        :param w2: w2 vessel measurement
        :type w2: float

        :return: the root vessel of w1 and w2 estimation
        :rtype: float
        """
        return 0.95 * (w1 ** 2 + w2 ** 2) ** 0.5

    def compute_central_retinal_equivalents(self,blood_vessel, skeleton, xc,yc, radius, artery = True, Toplot = False):
        """
        Compute the CRAE or CRVE equivalent for a given blood vessel graph.
        :param blood_vessel: blood_vessel segmentation containing binary values within {0,1}
        :type blood_vessel: np.array
        :param skeleton: blood_vessel segmentation skeleton containing binary values within {0,1}
        :type skeleton: np.array
        :param xc: x axis of the optic disc center
        :type xc: int
        :param yc: y axis of the optic disc center
        :type yc: int
        :param radius: radius in pixel of the optic disc
        :type radius: int
        :param artery: Flag to decide if to use CRAE or CRVE formulas (artery to True means CRAE, and to False means CRVE)
        :type artery: Bool
        :param Toplot: Flag to decide if to store the visualisation element. (Setting it to true use more RAM)
        :type Toplot: Bool


        :return: A tuple containing:
            - A result dictionnary (Dict): Dictionnary containing the computer CRE (-1 if it has failed)
            - plotable_list (List): A summary that contains the topology information required to plot the visualisation (really useful when Toplot is True). Return None if the computation has failed.
        :rtype: Tuple[Dict, List]
        """
        radius_zone_C = int(3 * radius)
        ## Extract the distances graphs
        B, D = self.extract_subgraphs(graphs=skeleton.copy(), x_c=xc, y_c=yc)

        ## Extract the starting points list by navigating through the skeleton graph
        starting_points = np.zeros((skeleton.shape[0], skeleton.shape[1]), dtype=float)
        for i in set(list(B.reshape(-1))) - {0}:
            mask = B == i
            if mask.sum() >= 50:
                min_index = (D * mask + (1 - mask) * 1e10).argmin()
                min_coordinates = np.unravel_index(min_index, D.shape)
                # print(((min_coordinates[0] - yc)**2 + (min_coordinates[1] - xc)**2)**0.5)
                if ((min_coordinates[0] - yc) ** 2 + (min_coordinates[1] - xc) ** 2) ** 0.5 < 20 + 2 * radius:
                    starting_points[min_coordinates[0], min_coordinates[1]] = 1
        starting_point_list = np.argwhere(starting_points == 1)


        ### Cleaning the skeleton graph irregularities
        B = np.zeros((blood_vessel.shape[0], blood_vessel.shape[1]))
        tree_reg_list = []
        plot = np.zeros((blood_vessel.shape[0], blood_vessel.shape[1]))
        for idx_start in starting_point_list:
            tree = TreeReg(idx_start[0], idx_start[1])
            tree.recursive_reg(skeleton.copy(), idx_start[0], idx_start[1], 0, tree, plot)
            tree_reg_list.append(tree)
        plot_list = []
        for tree_reg in tree_reg_list:
            plot = np.zeros((blood_vessel.shape[0], blood_vessel.shape[1]))
            tree_reg.print_reg(tree_reg, plot)
            plot_list.append(plot.copy())
        skoustideB_reg = np.sum(np.array(plot_list), axis=0)


        B = np.zeros_like(skeleton)
        D = np.zeros_like(skeleton)
        plotted_tmp = None if Toplot is False else np.zeros((blood_vessel.shape[0], blood_vessel.shape[1]))

        #####Initialising a dictionary that will be used to store the topology of th graph
        dico = {}

        ####Navigating through the graph to fill the the topology dico
        for idx_start in starting_point_list:
            i, j = idx_start
            tree = Tree((idx_start[0],idx_start[1]),(idx_start[0],idx_start[1]))
            tree.recursive_CRE(skeleton.copy(), B, D, idx_start[0], idx_start[1], 1, None, None, None, None, None,
                            None, None, None, None, None, None, None, plotted_tmp, i, j, tree, blood_vessel, xc, yc,
                            radius_zone_C)
            dico[(idx_start[0], idx_start[1])] = tree

        tree_list = list(dico.values())

        ####Second graph navigation to check consistency of the topology dico (parent vessel is supposed to be bigger than its children).
        final_list = []
        for i in range(len(tree_list)):
            tmp = tree_list[i].print_correct(tree_list[i], max_d=0)
            if type(tmp[2]) is not int:
                if tmp[0].startpoint != tmp[2].startpoint and np.mean(tmp[0].diameter_list) > np.mean(
                        tmp[2].diameter_list) + 2:
                    print('Potential error', i)
                    idx_start = tmp[0].startpoint
                    i, j = idx_start
                    tree = Tree((idx_start[0], idx_start[1]), (idx_start[0], idx_start[1]))
                    tree.recursive_CRE(skoustideB_reg.copy(), B, D, idx_start[0], idx_start[1], 1, None, None, None,
                                    None, None, None, None, None, None, None, None, None, plotted_tmp, i, j, tree, blood_vessel,
                                    xc, yc, radius_zone_C)
                    final_list.append(tree)  # .remove_children())
                else:
                    final_list.append(tmp[2])  # .remove_children())
            else:
                final_list.append(tmp[0])  # .remove_children())

        ### Aggregating the topology dico results
        plotable_list = []
        for tree in final_list:
            tree.plotable_show_tree(plotable_list, blood_vessel.shape, Toplot = Toplot)

        ### Compute and return the CRE VBMs
        if len(plotable_list) != 0:
            blood_vessel_lista = []
            blood_vessel_lista.append(plotable_list[0]['Mean diameter'])
            for j in range(1, len(plotable_list)):
                blood_vessel_lista.append(plotable_list[j]['Mean diameter'])

            if artery:
                craek = self.central_equivalent(np.sort(blood_vessel_lista)[-6:], self.crae_knudtson)
                craeh = self.central_equivalent(np.sort(blood_vessel_lista)[-6:], self.crae_hubbard)
                return ({"craek": craek, "craeh":craeh}, plotable_list)
            else:
                crvek = self.central_equivalent(np.sort(blood_vessel_lista)[-6:], self.crve_knudtson)
                crveh = self.central_equivalent(np.sort(blood_vessel_lista)[-6:], self.crve_hubbard)
                return ({"craek": crvek, "craeh": crveh}, plotable_list)
        else:
            if artery:
                return ({"craek": -1, "craeh": -1}, None)
            else:
                return ({"craek": -1, "craeh": -1}, None)